/*
 * Routines to draw Hershey Font characters
 *
 * By Eric Roskos, 10/30/88
 */
  
#include <stdio.h>

/* 
 * hchars.h is compiled with "compile.c" in this package.  The output
 * of compile.c is then edited to produce a font header module.
 */
   
#include "hchars.h"

/*
 * define DEBUG to display some size information on the fonts
 */
 
/* #define DEBUG		/* */

/*
 * The font size displayed is a function of n/d.  This is set
 * via the textscale(n, d) call.
 */
static int sn = 5, sd = 8;	/* scaling numerator and denominator */

/*
 * hchars points to an array of hershey font character definitions.
 * It is selected via the textfont(n) call.
 */
 static char **hchars = hchar500; /* current font selection */

/*
 * textlead determines how far down to space when a newline is seen
 * in a plotstr or hprintf call.  It is set via the textleading() call.
 */
static int textlead = 25;	/* inter-line leading for text */

/*
 * Debug variables
 */
#ifdef DEBUG
static int tog = 0; 		/* debugging toggle for width markers */
#endif

/*
 * _plotchar is the internal routine which draws a hershey font character
 * definition on the display.  The argument points to a hershey font
 * definition string.
 */
static void
_plotchar(pp)
char *pp;
{
int bx, by;	/* bx is dist from origin to left of char;
		   by is dist from origin to right of char */
int x, y;	/* the origin x and y coordinates for the character */
int nx, ny;	/* used to compute the x and y coord to move to */
int fmove = 1;	/* 1 if this is a move, 0 if it is a draw */
char *p = pp;	/* used to scan the character definition string */

	/* get current x and y coords on screen */
	x = getx();
	y = gety();

	/* get character dimensions */
	if (!p[0] || !p[1])
	{
		fprintf(stderr,"hershey string must be at least 2 chars\n");
		return;
	}
	
	bx = (((p[0]&0xff)-'R') * sn) / sd; /* dist from origin to left edge (neg)  */
	by = (((p[1]&0xff)-'R') * sn) / sd; /* dist from origin to right edge (pos) */

#ifdef DEBUG
	/* draw char dimensions */
	/* x dimension drawn w/an upward tick, y dimension a downward tick */
	/* alternate character dimensions are staggered vertically */
	moveto(x, y+7+(tog?0:6));
	lineto(x, y+10+(tog?0:6));
	lineto(x+bx, y+10+(tog?0:6));
	lineto(x+bx, y+7+(tog?0:6));
	moveto(x, y+10+(tog?0:6));
	lineto(x+by, y+10+(tog?0:6));
	lineto(x+by, y+13+(tog?0:6));
	moveto(x, y);
	tog = !tog;
	printf("bx=%d, by=%d\n", bx, by);
#endif

#ifndef DEBUG
	x -= (bx * sn) / sd ; /* make left edge of char start at user's origin */
#endif
	/* move past the dimensions */
	p += 2;
	
	/* draw the character itself */
	while (*p)
	{
		if (!p[1])
		{
			fprintf(stderr, "odd-length hershey char string: '%s'\n", pp);
			return;
		}
		
		/* space in p[0] means next coordinate should not draw */
		if (*p == ' ')
		{
			fmove++;
			p += 2;
			continue;	
		}
		
		/* compute scaled coordinate to move/draw to */
		nx = x + ((((p[0]&0xff)-'R') * sn) / sd);
		ny = y + ((((p[1]&0xff)-'R') * sn) / sd);
		
		/* now move or draw the next segment of the character */
		if (fmove)
			moveto(nx, ny);
		else
			lineto(nx, ny);
		
		/* go to next coordinate pair in string */
		p += 2;

		/* switch back to "draw" mode */
		fmove = 0;
	}

	/* move right to the next character position */
#ifdef DEBUG
	x += 20;
#else	
	x += by + 2;	/* dist from origin to right edge */
#endif
	moveto(x, y);
}	

/*
 * plotchar may be called with an ASCII character argument to draw
 * the argument character on the display at the current position.
 * The current position is set with the move() call.  Plotchar will
 * draw a box if the character is undefined.
 */
void
plotchar(c)
int c;
{
	if (hchars[c]==NULL)
		_plotchar(hchars[127]);
	else	
		_plotchar(hchars[c]);
}		

/*
 * plotstr may be called with a string argument to draw the entire
 * string on the display, starting at the current position.  If there
 * is a newline in the string, it will move down by textleading units,
 * return to the x coordinate at which this call to plotstr started,
 * and continue with the next character (if any).
 */
void
plotstr(s)
char *s;
{
int x, y;

	x = getx();
	y = gety();
	
	while (*s)
	{
		if (*s=='\n')
		{
			moveto(x, y+textlead);
			y += textlead;
		}
		else	
			plotchar(*s);
		s++;
	}
}		

/*
 * Printf to the display, using the hershey font
 */
void
hprintf(s,a,b,c,d,e,f,g,h,i,j,k,l,m,n)
char *s;
unsigned a,b,c,d,e,f,g,h,i,j,k,l,m,n;
{
char buf[256];

	sprintf(buf, s, a,b,c,d,e,f,g,i,j,k,l,m,n);
	plotstr(buf);
}	

/*
 * set the scaling numerator and denominator.  This determines the
 * size of the characters displayed.
 */
void
textscale(n, d)
int n, d;
{
	sn = n;
	sd = d;
}

/*
 * set the font to use.  The allowed arguments are 0..NHFONTS, where
 * NHFONTS is determined by what has been compiled into hchars.h.
 */
void
textfont(n)
int n;
{
	if (n >= NHFONTS)
	{
		fprintf(stderr, "textfont must be 0..%d\n", NHFONTS-1);
		return;
	}
	
	hchars = hfonts[n];
}	

/*
 * Set the vertical leading for newlines to the argument number of units.
 * The leading is not scaled by the scaling factor.  It is an absolute
 * number of "pixels", although pixels may be scaled by the plotting
 * routines used by this code.
 */
void
textleading(n)
int n;
{
	textlead = n;
}

#ifdef TEST

/*
 * sample test program
 */
main()
{
FILE *f;
int nr;
int c;
char defn[512];

	graphics();
	erasegraph();
	moveto(150, 150);
	textscale(2, 1);
	textfont(1);
	textleading(25);
	plotstr("This is a Hershey Test\n");
	hprintf("This is hprintf'ed %d\n", 12345);
	while (getchar()!='\n') ;
	text();
	exit(0);
}

#endif /* TEST */
