/* 
 * Copyright (C) 1986, 1987 Robert W. Berger
 * May be used for non-commercial purposes without prior written permission.
 * Functional enhancements by J. Eric Roskos, 1989, for TIROS-N tracking and
 * user-interface (see change log).
 */
 
/* Change Log
	3/11/89		v2.3B JER - Added plotting of orbits on graphics
			display.
			
	3/04/89		v2.3A JER - Adapted for NASA Prediction Bulletin format
			Added Unix-like command line, batch mode
	
	3/24/87		v2.3 Adapted for new kepler.dat format.
			     Allow beacon frequencies in mode.dat.
			     Use decay rate for drag compensation.
  
	5/10/86		v2.2 Added single character aliases for satellite
			names.
  
	4/30/86		v2.1 Print blank line if satellite dips below
			horizon and reappears during same orbit and day
  
	4/29/86		v2.0 Changed GetSatelliteParams() to use AMSAT's
			"kepler.dat" file. Moved schedule to "mode.dat" file.
  
        4/22/86         v1.3  Inserted N8FJB's suggestions for variable naming
                        which maintain 8 character uniqueness.
                        Also removed "include" file orbitr.h, which had two
                        definitions of external functions defined in orbit.c
			    -K3MC 
  
        4/1/86          v1.2  Corrected a scanf conversion to %d for an int
                        type.    -K3MC
 
        3/19/86         v1.1  Changed GetSatelliteParams to not pass NULL
                        to sscanf.
 
                                                                        */

/* 
 * Define PLOT to enable plotting of maps on display.
 * This option requires you to have both my (JER's) "maplib" software, and
 * the "The World Digitized" database (John B. Allison's shareware
 * database).  Given the complexities, you'd probably be better off
 * writing your own routine.  You just need the following:
 *
 * graphics()            - set the display into graphics mode
 * text()                - restore the display to normal text mode
 * plotmap()             - draw the base map of your area
 * mapmove(lat, lon)     - move cursor to (lat, lon) without drawing
 * mapline(lat, lon)     - move cursor to (lat, lon) drawing a line from
 *                         previous cursor position.
 * maplabel(fmt, <args>) - printf a label on the map at the current
 *                         cursor position.
 */

#define PLOT		/* */ 

/* Set DRAG to 1 to use drag info (none in NASA file, so not usable now */
#define DRAG 0

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <signal.h>

#define bzero(p, n) memset((char *)p, 0, n)

extern double   Kepler();
extern double   GetDay();


#define PI 3.14159265
#define PI2 (PI*2)
#define MinutesPerDay (24*60.0)
#define SecondsPerDay (60*MinutesPerDay)
#define HalfSecond (0.5/SecondsPerDay)
#define EarthRadius 6378.16		/* Kilometers           */
#define C 2.997925e5			/* Kilometers/Second    */
#define DegreesPerRadian (180/PI)
#define RadiansPerDegree (PI/180)
#define ABS(x) ((x) < 0 ? (-(x)) : (x))
#define SQR(x) ((x)*(x))

#define MaxModes 10
typedef struct
{
	int             MinPhase, MaxPhase;
	char            ModeStr[20];
}               ModeRec;

char            VersionStr[] = "N3EMO Orbit Simulator  v2.3B (NASA PB Format)";

 /* Keplerian Elements and misc. data for the satellite              */
double          EpochDay;		/* time of epoch                 */
double          EpochMeanAnomaly;		/* Mean Anomaly at epoch         */
long            EpochOrbitNum;	/* Integer orbit # of epoch      */
double          EpochRAAN;		/* RAAN at epoch                 */
double          epochMeanMotion;/* Revolutions/day               */
double          OrbitalDecay;	/* Revolutions/day^2             */
double          EpochArgPerigee;/* argument of perigee at epoch  */
double          Eccentricity;
double          Inclination;
char            SatName[100];
int             ElementSet;
double          BeaconFreq;		/* Mhz, used for doppler calc    */
double          MaxPhase;		/* Phase units in 1 orbit        */
double          perigeePhase;
int             NumModes;
ModeRec         Modes[MaxModes];
int             PrintApogee;

 /* Simulation Parameters */

double          StartTime, EndTime, StepTime;	/* In Days, 1 = New Year        */

 /* of reference year       */

 /* Site Parameters */
char            SiteName[100];
double          SiteLat, SiteLong, SiteAltitude, SiteMinElev;

/* other user input 	    */
int		Hr;				/* starting hour for simulation */
int		Month, Day, Year;		/* starting date */
double		duration;			/* duration of simulation */
char		stdate[20];			/* starting date */

/* option flags		    */
int 		verbose = 0;			/* verbose output */
int		prompt  = 0;			/* prompt for parameters */
int		sort	= 0;			/* produce "sortable" output */
int		plot	= 0;			/* plot on display */
int		world	= 0;			/* plot whole world map */

/* List the satellites in nasa.pb, and return the number found */
ListSatellites()
{
	char            str[100];
	FILE           *InFile;
	int             satnum;
	int             NumSatellites;
	int             lc = 0;

	if ((InFile = fopen("nasa.pb", "r")) == 0)
	{
		printf("\"nasa.pb\" not found\n");
		exit(-1);
	}

	if (fgets(str, 100, InFile)==NULL || 
	    strncmp(str, "- Current NASA", 14) != 0)
	{
		printf("Improper file format for file `nasa.pb'\n");
		printf("File should start with the line that gives the NASA bulletin number\n");
		exit(1);
	}
	
	if (!sort)
		printf("- Computations based on %s\n", str+10);

	if (prompt)
		printf("Available satellites:");

	satnum = 1;
	NumSatellites = 0;
	while (fgets(str, 100, InFile))
	{
		str[strlen(str)-1] = '\0';
		if ((lc++ % 3) != 0)
			continue;

		if (str[1] == ' ') /* heuristic */
		{
			printf("nasa.pb file out of sync: saw line %c when expected satellite name\n", str[0]);
			printf("Or, satellite name has a space for 2nd char\n");
			exit(1);
		}

		if (prompt)
		{
			if (((satnum-1) % 3) == 0) putchar('\n');
			printf("    %3d) %-15.15s", satnum++, str);
			if ((satnum % 58) == 0)
			{
				printf("\n[CR for More]");
				fgets(str, sizeof(str), stdin);
			}
		}
		NumSatellites++;
	}

	fclose(InFile);

	if (prompt)
		printf("\n\n");
	else
	if (verbose)
		printf("%d satellites in database\n", NumSatellites);	

	return NumSatellites;
}

/* Match and skip over a string in the input file. Exits on failure. */

MatchStr(InFile, FileName, Target)
FILE           *InFile;
char           *FileName, *Target;
{
	char            str[100];

	fgets(str, strlen(Target) + 1, InFile);
	if (strcmp(Target, str))
	{
		printf("%s: found \"%s\" while expecting \"%s\n\"", FileName, str, Target);
		exit(-1);
	}
}

/* Get an integer of digit length n */

int
GetStInt(s, n)
char *s;
int n;
{
char buf[80];
register char *p;

	bzero(buf, sizeof(buf));
	strncpy(buf, s, n);
	for (p=buf; *p; p++)	/* convert blanks to zeros */
		if (*p == ' ') *p = '0';
	return(atoi(buf));
}

/* get a floating point number of length n */

double
GetStDouble(s, n)
char *s;
int n;
{
char buf[80];
register char *p;

	bzero(buf, sizeof(buf));
	strncpy(buf, s, n);
	for (p=buf; *p; p++)	/* convert blanks to zeros */
		if (*p == ' ') *p = '0';
	return(atof(buf));
}	

/* return true if arg is a digit string, else return false */

int
IsANumber(s)
register char *s;
{
	while (*s)
	{
		if (*s < '0' || *s > '9') return(0);
		s++;
	}
	
	return(1);
}	

GetSatelliteParams()
{
	FILE           *InFile;
	char            str[100], str2[100];
	int             EpochYear;
	double          EpochHour, EpochMinute, EpochSecond;
	int             found;
	int             NumSatellites;
	int             satnum;
	int		ct;
	char	       *p;	

	NumSatellites = ListSatellites();

	found = 0;

	while (!found)
	{
		if (prompt)
		{
			printf("Number or satellite name :");
			gets(SatName);
		}

		if ((InFile = fopen("nasa.pb", "r")) == 0)
		{
			printf("nasa.pb not found\n");
			exit(-1);
		}

		fgets(str, 100, InFile);	/* skip header line */

		if (IsANumber(SatName))
		{						/* use single character label */
			satnum = atoi(SatName);
			if (satnum < 0 || satnum > NumSatellites)
			{
				printf("`%d' is out of range\n", satnum);
				fclose(InFile);
				exit(1);
			}

			while (--satnum > 0) /* skip previous entries */
			{
				if (fgets(str, 100, InFile) == NULL)
				{
				satnotf:
					printf("Satellite not found.\n");
					exit(1);
				}

				if (fgets(str, 100, InFile) == NULL)
					goto satnotf;
					
				if (fgets(str, 100, InFile) == NULL)
					goto satnotf;
			}
				
			fgets(str, 100, InFile); /* get satellite name line */
			p = &str[strlen(str) - 1];
			while (p >= str && (*p==' ' || *p=='\n'
				|| *p=='\t'))
			{
				*p-- = '\0';
			}

			found = 1;
		}

		else
		{
			while (!found)		/* use satellite name */
			{
				if (!fgets(str, 100, InFile))
					break;		/* EOF */

				p = &str[strlen(str) - 1];
				while (p >= str && (*p==' ' || *p=='\n'
					|| *p=='\t'))
				{
					*p-- = '\0';
				}

				if (strcmp(SatName, str) == 0)
					found = 1;
			}

			if (!found)
			{
				printf("Satellite `%s' not found\n", SatName);
				fclose(InFile);
				exit(1);
			}
		}
	}

	if (verbose)
		printf("Data for Satellite: %s\n\n", str);
	strcpy(SatName, str); /* this was a bug in the original prog - JER */
	
	BeaconFreq = 137.5;			/* Default value */

	/* get line 1 */
	fgets(str, 100, InFile);
	if (str[0] != '1')
	{
		printf("Bad file: first line numbered %c, not 1\n", str[0]);
		exit(1);
	}
	
	/* get line 2 */
	fgets(str2, 100, InFile);
	if (str2[0] != '2')
	{
		printf("Bad file: second line numbered %c, not 2\n", str[0]);
		exit(1);
	}
	
	/* Epoch Time */
	EpochDay = GetStDouble(&str[18], 14);
	if (verbose)
		printf("Epoch: ");

	EpochYear = EpochDay / 1000.0;
	EpochDay -= EpochYear * 1000.0;
	EpochDay = GetDay(EpochYear, 1, EpochDay);
	if (verbose)
	{
		PrintDate(stdout, EpochDay);
		putchar(' ');
		PrintTime(stdout, EpochDay);
		putchar('\n');
	}	

	ElementSet = GetStInt(&str[64], 4);
	if (verbose)
		printf("Element set %d\n", ElementSet);

	Inclination = GetStDouble(&str2[8], 8);
	if (verbose)
		printf("Inclination %lf\n", Inclination);
	Inclination *= RadiansPerDegree;

	EpochRAAN = GetStDouble(&str2[17], 8);
	if (verbose)
		printf("Right Ascension of Ascending Node: %lf\n", EpochRAAN);
	EpochRAAN *= RadiansPerDegree;

	str2[25] = '.';
	Eccentricity = GetStDouble(&str2[25], 8);
	if (verbose)
		printf("Eccentricity: %lf\n", Eccentricity);

	EpochArgPerigee = GetStDouble(&str2[34], 8);
	if (verbose)
		printf("Argument of Perigee: %lf\n", EpochArgPerigee);
	EpochArgPerigee *= RadiansPerDegree;

	EpochMeanAnomaly = GetStDouble(&str2[43], 8);
	if (verbose)
		printf("Mean Anomaly: %lf\n", EpochMeanAnomaly);
	EpochMeanAnomaly *= RadiansPerDegree;

	epochMeanMotion = GetStDouble(&str2[52], 11);
	if (verbose)
		printf("Mean Motion: %lf\n", epochMeanMotion);

#if DRAG /* no drag information in NASA file */
	MatchStr(InFile, "nasa.pb", "Decay rate:");
	fgets(str, 100, InFile);
	sscanf(str, "%lf", &OrbitalDecay);
#endif

	str[68] = '\0';
	EpochOrbitNum = atol(&str[63]);
	if (verbose)
		printf("Orbit number: %ld\n\n", EpochOrbitNum);

#if FREQS
	while (1)
	{
		if (!fgets(str, 100, InFile))
			break;				/* EOF */
		if (strlen(str) <= 2)
			break;				/* Blank line */
		sscanf(str, "Beacon: %lf", &BeaconFreq);
	}
#endif

	PrintApogee = (Eccentricity >= 0.3);

	perigeePhase = 0;
	MaxPhase = 360;				/* Default values */
	NumModes = 0;

	if ((InFile = fopen("mode.dat", "r")) == 0)
		return;

	found = 0;
	while (!found)
	{
		if (!fgets(str, 100, InFile))
			break;				/* EOF */
		str[strlen(str)-1] = '\0';
		if (strncmp(str, "Satellite:", 10) == 0)
		{
			if (strcmp(SatName, str+11) == 0)
				found = 1;
		}		
	}

	if (!found && !sort)
		printf("*** No mode.dat information found.\n");

	if (found)
	{
		while (1)
		{
			if (!fgets(str, 100, InFile))
				break;			/* EOF */
			if (strlen(str) <= 2)
				break;			/* Blank line */
			sscanf(str, "Beacon: %lf", &BeaconFreq);
			sscanf(str, "Perigee phase: %lf", &perigeePhase);
			sscanf(str, "Max phase: %lf", &MaxPhase);

			if (sscanf(str, "Mode: %20s from %d to %d", Modes[NumModes].ModeStr,
					&Modes[NumModes].MinPhase, &Modes[NumModes].MaxPhase) == 3
				&& NumModes < MaxModes)
				NumModes++;
		}
		fclose(InFile);
		
		if (verbose)
			printf("Transmitting frequency: %lf\n\n", BeaconFreq);
	}



}


GetSiteParams()
{
	FILE           *InFile;
	char            str[100];

#ifdef ASKSITE
	printf("Site name :");
	gets(str);
#else	/* use hardcoded local.sit */
	strcpy(str, "local");
#endif	
	strcat(str, ".sit");

	if ((InFile = fopen(str, "r")) == 0)
	{
		printf("%s not found\n", str);
		exit(-1);
	}

	fgets(SiteName, 100, InFile);

	fgets(str, 100, InFile);
	sscanf(str, "%lf", &SiteLat);
	SiteLat *= RadiansPerDegree;

	fgets(str, 100, InFile);
	sscanf(str, "%lf", &SiteLong);
	SiteLong *= RadiansPerDegree;

	fgets(str, 100, InFile);
	sscanf(str, "%lf", &SiteAltitude);
	SiteAltitude /= 1000;		/* convert to km */

	fgets(str, 100, InFile);
	sscanf(str, "%lf", &SiteMinElev);
	SiteMinElev *= RadiansPerDegree;
}

GetSimulationParams()
{
	double          hour;

	if (prompt)
	{
		printf("Start date (UTC) (Month/Day/Year): ");
		gets(stdate);
	}
	sscanf(stdate, "%d/%d/%d", &Month, &Day, &Year);
	if (verbose)
		printf("Start date: %d/%02d/19%02d\n", Month, Day, Year);

	StartTime = GetDay(Year, Month, (double) Day);

	if (prompt)
	{
		printf("Starting Time (UTC): ");
		scanf("%d", &Hr);
	}
	hour = (float)(Hr / 100);
	hour += (float)(Hr % 100) / 60.0;
	if (verbose)
		printf("Hour %lf, ", hour);
	StartTime += hour / 24;
	if (verbose)
		printf("computed start time: %lf\n", StartTime);

	if (prompt)
	{
		printf("Duration (Days):");
		scanf("%lf", &duration);
	}
	EndTime = StartTime + duration;

	if (prompt)
	{
		printf("Time Step (Minutes):");
		scanf("%lf", &StepTime);
	}
	StepTime /= MinutesPerDay;
}

PrintMode(OutFile, Phase)
FILE           *OutFile;

{
	int             CurMode;

	for (CurMode = 0; CurMode < NumModes; CurMode++)
		if ((Phase >= Modes[CurMode].MinPhase
				&& Phase <= Modes[CurMode].MaxPhase)
			|| ((Modes[CurMode].MinPhase > Modes[CurMode].MaxPhase)
				&& (Phase >= Modes[CurMode].MinPhase
					|| Phase <= Modes[CurMode].MaxPhase)))
		{
			fprintf(OutFile, "    %s", Modes[CurMode].ModeStr);
			break;
		}
}

/* fill in stdate with the default date string, and Hr with default hour */

DefDate()
{
long	tim;
struct tm *tm;

	time(&tim);
	tm = gmtime(&tim);
	
	sprintf(stdate, "%d/%02d/%02d", tm->tm_mon+1, tm->tm_mday,
		tm->tm_year % 100);
		
	Hr = tm->tm_hour*100 + tm->tm_min;
}

#ifdef PLOT
/*
 * Restore the display to normal in the event of a SIGINT
 */
terminate()
{
	text();
	exit(2);
}	
#endif

main(argc, argv)
int argc;
char **argv;
{
	double          ReferenceOrbit;		/* Floating point orbit # at epoch */
	double          CurrentTime;/* In Days                         */
	double          CurrentOrbit;
	double          AverageMotion,		/* Corrected for drag              */
	                CurrentMotion;
	double          MeanAnomaly, TrueAnomaly;
	double          SemiMajorAxis;
	double          Radius;		/* From geocenter                  */
	double          RAANPrecession, PerigeePrecession;
	double          SSPLat, SSPLong;
	long            OrbitNum, PrevOrbitNum;
	int             Day, PrevDay;
	double          Azimuth, Elevation, Range;
	double          PrevRange, Velocity, Doppler;
	int             Phase;
	char            FileName[100];
	FILE           *OutFile;
	int             DidApogee;
	double          TmpTime, PrevTime;
	int             PrevVisible;
	char		TimeBuf[25];
	float		PrevLongDeg = 0.0; /* used to detect map wraparound */
	int		FirstPoint = 1;    /* first point in this orbit */
	/*
	 * The following are used to label orbits on the whole-world map.
	 * We use PrevLat, LatDeriv, DerivSgn, and PrevDerivSgn to detect
	 * a pole crossing; then we label the LabPnt'th point in the orbit
	 * after that.  We then decrement LabPnt by one, for the next orbit.
	 * When LabPnt reaches zero, we reset it to FIRSTLAB.
	 */
#define FIRSTLAB 10
	int		LabPnt = FIRSTLAB, LabCnt;
	float		PrevLat = 0.0, LatDeriv;
	int		DerivSgn, PrevDerivSgn = 0;

	while (--argc > 0)
	{
		if (**++argv != '-') break;
		
		switch ((*argv)[1])
		{
		case 'p':
			plot++;
			break;
			
		case 'w':
			plot++;
			world++;
			break;

		case 'v':
			verbose++;
			break;
			
		case 'f':	/* fast, coarse-grained step */
			StepTime = 5.0;
			break;

		case 's':	/* "sortable" output */
			sort++;
			break;	
			
		default:
			printf("No such option as -%c\n", (*argv)[1]);
			printf("usage: orbit satname mm/yy/dd hhmm stepmins durdays\n");
			exit(1);
			break;
		}
	}

	if (!sort)
		printf("%s\n", VersionStr);

	/*
	 * set default step time
	 */
	if (world)
		StepTime = 5.0;
	else	
		StepTime = 1.0;

	/*
	 * get command line arguments
	 */
	if (argc <= 0)
		prompt++;

	if (argc > 0)
		strcpy(SatName, argv[0]);

	if (argc > 1)
		strcpy(stdate, argv[1]);
	else
		DefDate();	
		
	if (argc > 2)
		Hr = atoi(argv[2]);
		
	if (argc > 3)
		StepTime = atof(argv[3]);
		
	if (argc > 4)
		duration = atof(argv[4]);
	else
		duration = 1.0;	

	GetSatelliteParams();
	GetSiteParams();
	GetSimulationParams();

	FileName[0] = '\0';

	if (prompt)
	{
		printf("Output file (RETURN for TTY) :");
		gets(FileName);				/* Skip previous RETURN */
		gets(FileName);
	}

	if (strlen(FileName) > 0)
	{
		if ((OutFile = fopen(FileName, "w")) == 0)
		{
			printf("Can't write to %s\n", FileName);
			exit(-1);
		}
	}
	else
		OutFile = stdout;

	if (!sort)
	{
		fprintf(OutFile, "%s Element Set %d  ", SatName, ElementSet);
		fprintf(OutFile, "Antenna Site %s\n", SiteName);
		fprintf(OutFile, "Doppler calculated for freq = %lf MHz\n", BeaconFreq);
	}

#ifdef PLOT
	if (plot)	/* draw the base map on the display */
	{
		signal(SIGINT, terminate);
		graphics();
		erasegraph();
		if (world)
		{
			if (!plotmap("world.map", 45))
				text();
		}
		else
		{
			if (!plotmap("area.map", 15))
				text();
		}	

		mapmove(SiteLat*DegreesPerRadian, -SiteLong*DegreesPerRadian);
		maplabel("ANT");
	}
#endif

	SemiMajorAxis = 331.25 * exp(2 * log(MinutesPerDay / epochMeanMotion) / 3);
	GetPrecession(SemiMajorAxis, Eccentricity, Inclination, &RAANPrecession,
		&PerigeePrecession);

	ReferenceOrbit = EpochMeanAnomaly / PI2 + EpochOrbitNum;

	PrevDay = -10000;
	PrevOrbitNum = -10000;
	PrevTime = StartTime - 2 * StepTime;

	BeaconFreq *= 1E6;			/* Convert to Hz */

	DidApogee = 0;

	/* Start the loop one step early, to init OldRange */
	for (CurrentTime = StartTime - StepTime; CurrentTime <= EndTime;
		CurrentTime += StepTime)
	{

#if DRAG
		AverageMotion = epochMeanMotion
			+ (CurrentTime - EpochDay) * OrbitalDecay / 2;
		CurrentMotion = epochMeanMotion
			+ (CurrentTime - EpochDay) * OrbitalDecay;
#else
		AverageMotion = epochMeanMotion;
		CurrentMotion = epochMeanMotion;
#endif
		SemiMajorAxis = 331.25 * exp(2 * log(MinutesPerDay / CurrentMotion) / 3);

		CurrentOrbit = ReferenceOrbit +
			(CurrentTime - EpochDay) * AverageMotion;
		OrbitNum = CurrentOrbit;

		MeanAnomaly = (CurrentOrbit - OrbitNum) * PI2;

		TmpTime = CurrentTime;
		if (MeanAnomaly < PI)
			DidApogee = 0;
		if (PrintApogee && !DidApogee && MeanAnomaly > PI)
		{						/* Calculate Apogee */
			TmpTime -= StepTime;/* So we pick up later where we left off */
			MeanAnomaly = PI;
			CurrentTime = EpochDay + (OrbitNum - ReferenceOrbit + 0.5) / AverageMotion;
		}

		TrueAnomaly = Kepler(MeanAnomaly, Eccentricity);
		Radius = SemiMajorAxis * (1 - SQR(Eccentricity))
			/ (1 + Eccentricity * cos(TrueAnomaly));

		GetSubSatPoint(EpochDay, EpochRAAN, EpochArgPerigee,
			Inclination, Eccentricity, RAANPrecession, PerigeePrecession,
			CurrentTime, TrueAnomaly, &SSPLat, &SSPLong);
		GetBearings(SiteLat, SiteLong, SiteAltitude, SSPLat, SSPLong, Radius,
			&Azimuth, &Elevation, &Range);

		Velocity = (Range - PrevRange) / ((CurrentTime - PrevTime) * SecondsPerDay);
		PrevRange = Range;

#ifdef MSDOS
		kbhit(); 	/* allow user to cancel or can run long time */
#endif			

		if ((world || Elevation >= SiteMinElev) && 
			CurrentTime >= StartTime)
		{
			Day = CurrentTime + HalfSecond;
			if (((double) Day) > CurrentTime + HalfSecond)
				Day -= 1;		/* Correct for truncation of negative values */

			if (OrbitNum == PrevOrbitNum && Day == PrevDay && !PrevVisible)
				fprintf(OutFile, "\n");	/* Dipped out of sight; print blank */
#ifdef PLOT
			/* pen-up at start of new orbit */
			if (!world && OrbitNum != PrevOrbitNum)
			{
				FirstPoint = 1;
			}

			/* 
			 * label this orbit at LabPnt points past
			 * pole crossing; LabCnt counts how
			 * many points we've plotted since then.
			 */
			LatDeriv = SSPLat - PrevLat;
			PrevLat = SSPLat;
			DerivSgn = (LatDeriv > 0.0? 1 : -1);
			if (PrevDerivSgn && PrevDerivSgn != DerivSgn)
			{
				LabCnt = 0;
			}
			PrevDerivSgn = DerivSgn;
#endif
			if (!sort && (OrbitNum != PrevOrbitNum || Day != PrevDay))
			{					/* Print Header */
				PrintDate(OutFile, Day);
				fprintf(OutFile, "  ----%s Orbit # %ld-----\n", SatName, OrbitNum);
				fprintf(OutFile, " U.T.C.   Az   El  Doppler  Range");
				fprintf(OutFile, "  Height  Lat  Long  Phase(%3.0lf)\n",
					MaxPhase);
			}
			PrevOrbitNum = OrbitNum;
			PrevDay = Day;
			if (sort)
				PrintXTime(OutFile, Day, CurrentTime + HalfSecond);
			else	
				PrintTime(OutFile, CurrentTime + HalfSecond);
#ifdef PLOT
			if (plot) /* get time ready for labelling point */	
			{
				SPrintXTime(TimeBuf, Day, 
					    CurrentTime + HalfSecond);
			}	
#endif
			Doppler = -BeaconFreq * Velocity / C;

			fprintf(OutFile, "  %3.0lf  %3.0lf  %6.0lf  %6.0lf",
				Azimuth * DegreesPerRadian,
				Elevation * DegreesPerRadian, Doppler, Range);

			fprintf(OutFile, "  %6.0lf  %3.0lf  %4.0lf",
				Radius - EarthRadius, SSPLat * DegreesPerRadian,
				SSPLong * DegreesPerRadian);
#ifdef PLOT
			if (plot)
			{
			float LongTmp;
				
				/* 
				 * convert ORBIT23's coordinates to map coords;
				 * detect if we're wrapping-around the map.
				 */
				LongTmp = -SSPLong * DegreesPerRadian;
				if (LongTmp < -180.0)
					LongTmp = 360.0 + LongTmp;

				if ((PrevLongDeg < 0.0 && LongTmp > 0.0) ||
				    (PrevLongDeg > 0.0 && LongTmp < 0.0))
					FirstPoint = 1;
					
				PrevLongDeg = LongTmp;

				/* plot the subpoint */
				
				if (FirstPoint)
				{
					mapmove(SSPLat * DegreesPerRadian,
						LongTmp);
					FirstPoint = 0;	
				}
				else
				{
					mapline(SSPLat * DegreesPerRadian,
						LongTmp);
				}

				/* label the subpoint */
				
				if (!world || LabPnt == ++LabCnt)
				{
					maplabel("%sZ %ld", TimeBuf, OrbitNum);
					LabPnt--;
					if (LabPnt <= 0) LabPnt = FIRSTLAB;
				}	
			}
#endif
			Phase = (MeanAnomaly / PI2 * MaxPhase + perigeePhase);
			while (Phase < 0)
				Phase += MaxPhase;
			while (Phase >= MaxPhase)
				Phase -= MaxPhase;

			fprintf(OutFile, "  %4d", Phase);
			PrintMode(OutFile, Phase);
			if (sort)
				fprintf(OutFile, "  %15.15s", SatName);
			if (PrintApogee && (MeanAnomaly == PI))
				fprintf(OutFile, "    Apogee");
			fprintf(OutFile, "\n");
			PrevVisible = 1;
		}
		else
			PrevVisible = 0;
		if (PrintApogee && (MeanAnomaly == PI))
			DidApogee = 1;
		PrevTime = CurrentTime;
		CurrentTime = TmpTime;
	}
	fclose(OutFile);

#ifdef PLOT
	if (plot)
	{
		printf("Press [Return] to continue...\007");
		while (getchar() != '\n') ;
	}
#endif	
}
